#include"k3menu.h"

#include<string.h>
#include<stdio.h>
#include<GLFW/glfw3.h>

void k3MenuSetBounds_(struct k3MObj *this, int16_t x, int16_t y, int16_t w, int16_t h) {
	this->x = x;
	this->y = y;
	this->w = w;
	this->h = h;
}

static bool event_send_nonrecur(struct k3MEvent *ev) {
	bool handled = false;
	for(size_t i = 0; i < ev->target->handlerCount; i++) {
		struct k3MEventHandler *h = &ev->target->handlers[i];
		
		if(h->code == ev->code || h->code == k3M_EVENT_ALL) {
			if(h->callback(ev, h->ud)) {
				return true;
			}
		}
	}
	
	return handled;
}

int k3MRegisterEventHandler_(struct k3MObj *obj, uint16_t evcode, k3MEventHandlerFunc callback, uint8_t *ud, size_t udSize) {
	if(ud && udSize > k3M_USERDATA_SIZE) {
		return 0;
	}
	
	obj->handlers = realloc(obj->handlers, sizeof(*obj->handlers) * ++obj->handlerCount);
	
	obj->handlers[obj->handlerCount - 1].code = evcode;
	obj->handlers[obj->handlerCount - 1].callback = callback;
	if(ud) {
		memcpy(obj->handlers[obj->handlerCount - 1].ud, ud, udSize);
	}
}

bool k3MEventSend(struct k3MEvent *ev) {
	while(!event_send_nonrecur(ev)) {
		if(ev->kind == k3M_EVENTKIND_DIRECT) {
			return false;
		} else if(ev->kind == k3M_EVENTKIND_BUBBLE) {
			ev->target = ev->target->parent;
			
			if(ev->target == ev->original || !ev->target) {
				return false;
			}
		} else if(ev->kind == k3M_EVENTKIND_CAPTURE) {
			struct k3MObj *parent = ev->target;
			
			for(size_t i = 0; i < parent->childCount; i++) {
				if(!parent->children[i]->invisible) {
					ev->target = parent->children[i];
					
					if(k3MEventSend(ev)) {
						return true;
					}
				}
			}
			
			return false;
		}
	}
	
	return true;
}

static bool label_draw(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MLabel *this = (void*) ev->target;
	
	if(this->txt) {
		k3FontDraw(this->font, this->x, this->y, this->sz, this->txt, (vec4) {1, 1, 1, 1});
	}
	
	return false;
}

struct k3MLabel *k3MLabel(struct k3Font *font, float sz, const char *txt) {
	struct k3MLabel *ret = calloc(1, sizeof(*ret));
	ret->font = font;
	ret->sz = sz;
	ret->txt = txt;
	
	k3MRegisterEventHandler(ret, k3M_EVENT_DRAW, label_draw, NULL, 0);
	
	return ret;
}

int k3MRemoveChild(struct k3MObj *parent, struct k3MObj *child) {
	for(size_t i = 0; i < parent->childCount; i++) {
		if(parent->children[i] == child) {
			memcpy(&parent->children[i], &parent->children[i + 1], sizeof(*parent->children) * (parent->childCount - i - 1));
			
			parent->childCount--;
			
			child->parent = NULL;
			
			return 1;
		}
	}
	
	return 0;
}

void k3MAddChild(struct k3MObj *parent, struct k3MObj *child) {
	if(child->parent) {
		k3MRemoveChild(child->parent, child);
	}
	
	parent->children = realloc(parent->children, sizeof(*parent->children) * (++parent->childCount));
	parent->children[parent->childCount - 1] = child;
	
	child->parent = parent;
}

/*static struct k3Menu *k3ScreenFuncMouse(struct k3Menu *_, int16_t x, int16_t y, int ev) {
	struct k3Screen *this = (void*) _;
	
	if(ev == k3_MENU_MOUSE_MOVE) {
		struct k3Menu *newHovered = k3ContainerFuncMouse(_, x, y, k3_MENU_MOUSE_TEST);
		if(this->hovered != newHovered) {
			if(this->hovered && this->hovered->funcmouse) {
				this->hovered->funcmouse(this->hovered, x, y, k3_MENU_MOUSE_OUT);
			}
			
			this->hovered = newHovered;
			
			if(this->hovered && this->hovered->funcmouse) {
				this->hovered->funcmouse(this->hovered, x, y, k3_MENU_MOUSE_IN);
			}
		} else {
			if(this->hovered && this->hovered->funcmouse) {
				return this->hovered->funcmouse(this->hovered, x, y, k3_MENU_MOUSE_MOVE);
			}
		}
	} else if(this->hovered && this->hovered->funcmouse) {
		struct k3Menu *ret = this->hovered->funcmouse(this->hovered, x, y, ev);
		this->keyboardFocus = ret;
		return ret;
	}
	
	return NULL;
}

static void k3ScreenFuncKey(struct k3Menu *_, int key, int action, int modifiers) {
	struct k3Screen *this = (void*) _;
	
	if(this->keyboardFocus && this->keyboardFocus->funckey) {
		this->keyboardFocus->funckey(this->keyboardFocus, key, action, modifiers);
	}
}

static void k3ScreenFuncChar(struct k3Menu *_, uint32_t codepoint) {
	struct k3Screen *this = (void*) _;
	
	if(this->keyboardFocus && this->keyboardFocus->funcchar) {
		this->keyboardFocus->funcchar(this->keyboardFocus, codepoint);
	}
}*/

static bool screen_ev(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MScreen *this = (void*) ev->target;
	
	if(ev->code == k3M_EVENT_MOUSE_PRESS || ev->code == k3M_EVENT_MOUSE_RELEASE || ev->code == k3M_EVENT_MOUSE_MOTION) {
		struct k3MObj *innermost = (void*) this;
		
		while(innermost->childCount != 0) {
			
			bool foundInner = false;
			
			for(intmax_t i = innermost->childCount - 1; i >= 0; i--) {
				
				struct k3MObj *child = innermost->children[i];
				
				if(!child->invisible && ev->mouse.x >= child->x && ev->mouse.x < child->x + child->w && ev->mouse.y >= child->y && ev->mouse.y < child->y + child->h) {
					
					innermost = child;
					foundInner = true;
					
					break;
					
				}
				
			}
			
			if(!foundInner) {
				break;
			}
			
		}
		
		if(innermost != (void*) this) {
			
			if(ev->code == k3M_EVENT_MOUSE_PRESS) {
				this->keyboardFocus = innermost;
			}
			
			ev->kind = k3M_EVENTKIND_BUBBLE;
			ev->target = innermost;
			k3MEventSend(ev);
			
			if(ev->code == k3M_EVENT_MOUSE_RELEASE) {
				ev->code = k3M_EVENT_MOUSE_CLICK;
				
				ev->kind = k3M_EVENTKIND_DIRECT;
				ev->target = innermost;
				k3MEventSend(ev);
			}
			
		}
		
		return true;
	} else if(ev->code == k3M_EVENT_KEY_PRESS || ev->code == k3M_EVENT_KEY_RELEASE || ev->code == k3M_EVENT_INPUT) {
		
		if(this->keyboardFocus) {
			ev->kind = k3M_EVENTKIND_BUBBLE;
			ev->target = (void*) this->keyboardFocus;
			k3MEventSend(ev);
		}
		
		return true;
		
	}
	
	ev->kind = k3M_EVENTKIND_CAPTURE;
	for(intmax_t i = this->childCount - 1; i >= 0; i--) {
		if(!this->children[i]->invisible) {
			ev->target = this->children[i];
			if(k3MEventSend(ev)) {
				return true;
			}
		}
	}
	
	return false;
}
struct k3MScreen *k3MScreen() {
	struct k3MScreen *ret = calloc(1, sizeof(*ret));
	
	k3MRegisterEventHandler(ret, k3M_EVENT_ALL, screen_ev, NULL, 0);
	
	return ret;
}

static bool textbutton_draw(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MTextButton *this = (void*) ev->target;
	
	if(this->txt) {
		k3BatchAdd(NULL, (struct k3RectF) {}, (struct k3RectF) {
			this->x, this->y,
			this->w, this->h
		}, 0, (vec4) {1, 1, 1, 0.2 + 0.1 * this->hovered});
		
		struct k3RectF txtsz;
		k3FontSz(this->font, this->sz, this->txt, &txtsz);
		
		k3FontDraw(this->font, this->x + this->w / 2 - txtsz.w / 2, this->y, this->sz, this->txt, (vec4) {
			1 - 0.8 * this->disabled, 1 - 0.8 * this->disabled, 1 - 0.8 * this->disabled, 1});
	}
	
	return false;
}
struct k3MTextButton *k3MTextButton(struct k3Font *font, float sz, const char *txt) {
	struct k3MTextButton *ret = calloc(1, sizeof(*ret));
	
	ret->font = font;
	ret->sz = sz;
	ret->txt = txt;
	
	k3MRegisterEventHandler(ret, k3M_EVENT_DRAW, textbutton_draw, NULL, 0);
	
	return ret;
}

static bool textinput_draw(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MTextInput *this = (void*) ev->target;
	
	if(this->txt) {
		k3BatchAdd(NULL, (struct k3RectF) {}, (struct k3RectF) {
			this->x, this->y,
			this->w, this->h
		}, 0, (vec4) {1, 1, 1, 0.2});
		
		int isPlaceholder = strlen(this->txt) == 0;
		const char *txt = isPlaceholder ? this->placeholder : this->txt;
		
		struct k3RectF txtsz;
		k3FontSz(this->font, this->sz, txt, &txtsz);
		
		k3FontDraw(this->font, this->x, this->y, this->sz, txt, (vec4) {1 - 0.5 * isPlaceholder, 1 - 0.5 * isPlaceholder, 1 - 0.5 * isPlaceholder, 1});
	}
	
	return false;
}
static bool textinput_key(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MTextInput *this = (void*) ev->target;
	
	if(ev->key.num == GLFW_KEY_BACKSPACE && ev->code != k3M_EVENT_KEY_RELEASE) {
		char *last = k3UTF8LastCodepointZ(this->txt);
		if(last) {
			*last = 0;
		}
	} else if(ev->key.num == GLFW_KEY_ENTER && ev->code == k3M_EVENT_KEY_PRESS) {
		struct k3MEvent newev = {
			.code = k3M_EVENT_COMPLETE,
			.kind = k3M_EVENTKIND_DIRECT,
			
			.original = (void*) this,
			.target = (void*) this,
		};
		
		k3MEventSend(&newev);
	}
	
	return true;
}
static bool textinput_char(struct k3MEvent *ev, uint8_t *ud) {
	struct k3MTextInput *this = (void*) ev->target;
	
	size_t len = strlen(this->txt);
	
	this->txt = realloc(this->txt, len + 5);
	len += k3UTF8Encode(ev->input.code, this->txt + len);
	this->txt[len] = 0;
	
	return true;
}
struct k3MTextInput *k3MTextInput(struct k3Font *font, float sz, const char *placeholder, const char *txt) {
	struct k3MTextInput *ret = calloc(1, sizeof(*ret));
	
	ret->font = font;
	ret->sz = sz;
	ret->placeholder = placeholder ? placeholder : "";
	ret->txt = strdup(txt ? txt : "");
	
	k3MRegisterEventHandler(ret, k3M_EVENT_DRAW, textinput_draw, NULL, 0);
	k3MRegisterEventHandler(ret, k3M_EVENT_KEY_PRESS, textinput_key, NULL, 0);
	k3MRegisterEventHandler(ret, k3M_EVENT_KEY_RELEASE, textinput_key, NULL, 0);
	k3MRegisterEventHandler(ret, k3M_EVENT_INPUT, textinput_char, NULL, 0);
	
	return ret;
}

struct k3MObj *k3MObj() {
	struct k3MObj *ret = calloc(1, sizeof(*ret));
	return ret;
}
